/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// Ptloc.cpp    Implement fast point location routines.                    //
//                                                                         //
// by Si hang March, 2001.                                                 //
// stsc_sh@sina.com                                                        //
//                                                                         //
// This file with it's couple file ptloc.cpp were specially written for    //
//   the FEMM program(http://groups.yahoo.com/group/femm/) femmview to     //
//   provide a fast point location function to improve the Line Integral   //
//   caculation speed when the mesh size is big.                           //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <afx.h>
#include <afxtempl.h>
#include "complex.h"      // declared CComplex
#include "problem.h"      // declared CMeshNode, CMeshElem 

#include "ptloc.h"
#include <math.h>

// A large value to indicate a impossible element index.
#define INVALIDELEINDEX    0x7fffffff
// We use the highest bit of a integer as boundary mark.
#define BOUNDFLAG          0x80000000

// some macros for convenience

#define TriEdge_Div4  >> 2
#define TriEdge_Mul4  << 2
#define TriEdge_Mod4  & 03

#define Eleindex(E)   ( (E) TriEdge_Div4 )
#define Orient(E)     ( (E) TriEdge_Mod4 )
#define Encode(T, V)  ( ((T) TriEdge_Mul4) + (V) )

// NOTE: Above 3 macros act like functions!

///////////////////////////////////////////////////////////////////////////////
// Local variables declaration                                               //
// BEGIN                                                                     //
///////////////////////////////////////////////////////////////////////////////

// These two variables should be init before use.

static CArray< CMeshNode, CMeshNode&> *pmeshnode;
static CArray< CElement, CElement&>   *pmeshelem;

// Fast lookup arrays to speed some of the mesh manipulation primitives.

static int plus1mod3[3] = {1, 2, 0};
static int minus1mod3[3] = {2, 0, 1};

// Variables for fast locate point.

// Keep a recently visited triangle's index.  Improves point location
//   if proximate points are inserted sequentially.
static TriEdge recenttri;

// Number of random samples for point location.
static int samples;

// Current random number seed.   
static unsigned long randomseed;  

///////////////////////////////////////////////////////////////////////////////
// END                                                                       //
// Local variables declaration                                               //
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// class TriEdge implementation                                              //
// BEGIN                                                                     //
///////////////////////////////////////////////////////////////////////////////

TriEdge::TriEdge() : ele(INVALIDELEINDEX), ori(0)
{
}

TriEdge::TriEdge(int _ele, int _ori) : ele(_ele), ori(_ori)
{
}

TriEdge::TriEdge(TriEdge& src) : ele(src.ele), ori(src.ori)
{
}

TriEdge& TriEdge::operator=(const TriEdge& src)
{
  ele = src.ele;
  ori = src.ori;
  return *this;
}

BOOL TriEdge::operator==(TriEdge& src)
{
  return (ele == src.ele) && (ori == src.ori);
}

BOOL TriEdge::operator!=(TriEdge& src)
{
  return (ele != src.ele) || (ori != src.ori);
}

TriEdge TriEdge::sym()
{
  int E = ((*pmeshelem)[ele]).n[ori];
  return TriEdge(Eleindex(E), Orient(E));
}

TriEdge& TriEdge::symself()
{
  int E = ((*pmeshelem)[ele]).n[ori];
  ele = Eleindex(E);
  ori = Orient(E);
  return *this;
}

TriEdge TriEdge::lnext()
{
  return TriEdge(ele, plus1mod3[ori]);
}

TriEdge& TriEdge::lnextself()
{
  ori = plus1mod3[ori];
  return *this;
}

TriEdge TriEdge::lprev()
{
  return TriEdge(ele, minus1mod3[ori]);
}

TriEdge& TriEdge::lprevself()
{
  ori = minus1mod3[ori];
  return *this;
}

TriEdge TriEdge::onext()
{
  return lprev().symself();
}

TriEdge& TriEdge::onextself()
{
  lprevself().symself();
  return *this;
}

TriEdge TriEdge::oprev()
{
  return sym().lnextself();
}

TriEdge& TriEdge::oprevself()
{
  symself().lnextself();
  return *this;
}

TriEdge TriEdge::dnext()
{
  return sym().lprevself();
}

TriEdge& TriEdge::dnextself()
{
  symself().lprevself();
  return *this;
}

TriEdge TriEdge::dprev()
{
  return lnext().symself();
}

TriEdge& TriEdge::dprevself()
{
  lnextself().symself();
  return *this;
}

TriEdge TriEdge::rnext()
{
  return sym().lnextself().symself();
}

TriEdge& TriEdge::rnextself()
{
  symself().lnextself().symself();
  return *this;
}

TriEdge TriEdge::rprev()
{
  return sym().lprevself().symself();
}

TriEdge& TriEdge::rprevself()
{
  symself().lprevself().symself();
  return *this;
}

CMeshNode* TriEdge::org()
{
  int N = ((*pmeshelem)[ele]).p[plus1mod3[ori]]; 
  if(N < 0 || N >= pmeshnode->GetSize()) throw;	
  return &((*pmeshnode)[N]);
}

CMeshNode* TriEdge::dest()
{
  int N = ((*pmeshelem)[ele]).p[minus1mod3[ori]]; 
  if(N < 0 || N >= pmeshnode->GetSize()) throw;	
  return &((*pmeshnode)[N]);
}

CMeshNode* TriEdge::apex()
{
  int N = ((*pmeshelem)[ele]).p[ori];  
  if(N < 0 || N >= pmeshnode->GetSize()) throw;	
  return &((*pmeshnode)[N]);
}

BOOL TriEdge::symexist()
{
  int E = ((*pmeshelem)[ele]).n[ori];
  return !(E & BOUNDFLAG);
}

void TriEdge::bondboundary(TriEdge& tedge)
{
  // Here tedge must also be a boundary edge.
  if(tedge.symexist()) throw;
  // This edge's neigh must be Outer Space.
  int E = ((*pmeshelem)[ele]).n[ori];
  if(!(E & BOUNDFLAG)) throw;
  // Now bound tedge as this edge's neigh.
  // Note: We also need keep the BOUNDFLAG in its neigh.
  E = Encode(tedge.ele, tedge.ori);
  E |= BOUNDFLAG;
  ((*pmeshelem)[ele]).n[ori] = E;
}

TriEdge TriEdge::nextboundary()
{
	// Here this TriEdge must be a boundary edge.
  if(symexist()) throw;
  // This edge's neigh must be Outer Space.
  int E = ((*pmeshelem)[ele]).n[ori];
  if(!(E & BOUNDFLAG)) throw;
  // Get the true TriEdge handle.
  // Note: We must get off the BOUNDFLAG from its neigh first.
  E = (E & ~BOUNDFLAG);
  return TriEdge(Eleindex(E), Orient(E));
}

TriEdge& TriEdge::nextboundaryself()
{
	// Here this TriEdge must be a boundary edge.
  if(symexist()) throw;
  // This edge's neigh must be Outer Space.
  int E = ((*pmeshelem)[ele]).n[ori];
  if(!(E & BOUNDFLAG)) throw;
  // Get the true TriEdge handle.
  // Note: We must get off the BOUNDFLAG first.
  E = (E & ~BOUNDFLAG);
  ele = Eleindex(E);
  ori = Orient(E);
  return *this;
}
///////////////////////////////////////////////////////////////////////////////
// END                                                                       //
// class TriEdge implementation                                              //
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Geometry Predicates functions                                             //
// BEGIN                                                                     //
///////////////////////////////////////////////////////////////////////////////

// Compare 2 floating point values dx and dy.
//   return +1 if dx > dy
//   return -1 if dx < dy
//   return  0 if dx == dy       
int FuzzyComp(const double dx, const double dy)
{
  double dSum  = fabs(dx) + fabs(dy);
  double dDiff = dx - dy;
  const double dTol = 5.e-9;
  const double dFloor = 1.e-6;

  dSum = (dSum > dFloor) ? dSum : dFloor;
  if (dDiff > dTol * dSum) return (1);
  else if (dDiff < - dTol * dSum) return (-1);
  else return (0);
}

// Orientation 2D, input 3 points dA, dB, dC.
//   return +1 if they are occur in counterclockwise order.
//   return -1 if they are occur in clockwise order.
//   return  0 if they are collinear. 
int Orient2D(const double dAx, const double dAy, 
             const double dBx, const double dBy,
             const double dCx, const double dCy)
{             
  // Currently no exact arithmetic
  double dDet = ((dBx - dAx) * (dCy - dAy) - (dCx - dAx) * (dBy - dAy));
  // Scale the determinant by the mean of the magnitudes of vectors:
  double distAB = sqrt((dAx - dBx) * (dAx - dBx) + (dAy - dBy) * (dAy - dBy)),
         distBC = sqrt((dBx - dCx) * (dBx - dCx) + (dBy - dCy) * (dBy - dCy)),
         distCA = sqrt((dCx - dAx) * (dCx - dAx) + (dCy - dAy) * (dCy - dAy));
  double dScale = (distAB + distBC + distCA) / 3; 
  dDet /= (dScale * dScale);
  if (dDet > 1.e-10) return 1;
  else if (dDet < -1.e-10) return -1;
  else return 0;
}

int Orient2D(CMeshNode& dA, CMeshNode& dB, CMeshNode& dC)
{
  return Orient2D(dA.x, dA.y, dB.x, dB.y, dC.x, dC.y);
}

///////////////////////////////////////////////////////////////////////////////
// END                                                                       //
// Geometry Predicates functions                                             //
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Ptloc Preprocess Functions                                                //
// BEGIN                                                                     //
///////////////////////////////////////////////////////////////////////////////

void BuildOuterBoundaryLink(TriEdge& startedge)
{
  TriEdge thisedge, nextedge;

  thisedge = startedge;
  // Get startedge's next edge in the same element.
  nextedge = thisedge.lnext();  
  while(TRUE) {
    // Spin nextedge around its org until it encounter an Outer Boundary.
    while(nextedge.symexist()) { 
      nextedge.oprevself();  
    }
    // Add nextedge to link(after thisedge).
    thisedge.bondboundary(nextedge);
    // Check to see if we can leave.
    if(nextedge == startedge) break;
    thisedge = nextedge;
    nextedge = thisedge.lnext();
  }
}

void BuildElemMap(CArray< CMeshNode, CMeshNode&> *pMeshNodes,
                  CArray< CElement, CElement&> *pMeshElems,
                  int *pNumList, int **ppConList)
{
  // Init variables for Point location functions.
  recenttri.ele = INVALIDELEINDEX;
  samples = 1; 
  randomseed = 1;
  // Init pmeshnode, pmeshelem.
  pmeshnode = pMeshNodes;
  pmeshelem = pMeshElems;
  	
  int i, j;

  // Init all elements' neigh to be unfinished.
  for(i = 0; i < pmeshelem->GetSize(); i ++) {
    for(j = 0; j < 3; j ++)
	  ((*pmeshelem)[i]).n[j] = INVALIDELEINDEX;
  }
  
  int orgi, desti;
  int ei, ni;
  BOOL done;
  // Loop all elements, to find and set there neighs.
  for(i = 0; i < pmeshelem->GetSize(); i ++) {
    for(j = 0; j < 3; j ++) {
      if(((*pmeshelem)[i]).n[j] == INVALIDELEINDEX) {
        // Get this edge's org and dest node index, 
        orgi = ((*pmeshelem)[i]).p[plus1mod3[j]];
        desti = ((*pmeshelem)[i]).p[minus1mod3[j]];
        done = FALSE;
        // Find this edge's neigh from the org node's list
        for(ni = 0; ni < pNumList[orgi]; ni ++) {
          // Find a Element around org node contained dest node of this edge. 
          ei = ppConList[orgi][ni];
          if(ei == i) continue; // Skip myself.
          // Check this Element's 3 vert to see if there exist dest node.
          if(((*pmeshelem)[ei]).p[0] == desti) {
            if(((*pmeshelem)[ei]).p[1] == orgi) {
              ((*pmeshelem)[i]).n[j] = Encode(ei, 2);
              ((*pmeshelem)[ei]).n[2] = Encode(i, j);
            } else if (((*pmeshelem)[ei]).p[2] == orgi) {   
              ((*pmeshelem)[i]).n[j] = Encode(ei, 1); 
              ((*pmeshelem)[ei]).n[1] = Encode(i, j);
            } else {
              throw; // Shouldn't be here.
            }
            done = TRUE;
            break;
          } else if(((*pmeshelem)[ei]).p[1] == desti) {
            if(((*pmeshelem)[ei]).p[2] == orgi) {
              ((*pmeshelem)[i]).n[j] = Encode(ei, 0);
              ((*pmeshelem)[ei]).n[0] = Encode(i, j);
            } else if(((*pmeshelem)[ei]).p[0] == orgi) {   
              ((*pmeshelem)[i]).n[j] = Encode(ei, 2); 
              ((*pmeshelem)[ei]).n[2] = Encode(i, j);
            } else {
              throw; // Shouldn't be here.
            }
            done = TRUE;
            break;
          } else if(((*pmeshelem)[ei]).p[2] == desti) {
            if(((*pmeshelem)[ei]).p[0] == orgi) {
              ((*pmeshelem)[i]).n[j] = Encode(ei, 1);
              ((*pmeshelem)[ei]).n[1] = Encode(i, j);
            } else if(((*pmeshelem)[ei]).p[1] == orgi) {  
              ((*pmeshelem)[i]).n[j] = Encode(ei, 0); 
              ((*pmeshelem)[ei]).n[0] = Encode(i, j);
            } else {
              throw; // Shouldn't be here.
            }
            done = TRUE;
            break;
          } 
        }
        if(!done) {
          // This edge must be a Boundary Edge.
          ((*pmeshelem)[i]).n[j] |= BOUNDFLAG;
        }
      } // Finish One Edge        
    } // End of One Element Loop
  } // End of Main Loop
  
  // Build links for Outer Boundary. These links used for fast point 
  //   location functions.
  for( i = 0; i < pmeshelem->GetSize(); i ++ ) {
    for( j = 0; j < 3; j++ ) {
      if(((*pmeshelem)[i]).n[j] & BOUNDFLAG) {
      	if((((*pmeshelem)[i]).n[j] & ~BOUNDFLAG) == INVALIDELEINDEX) {
          TriEdge startedge(i, j);
          BuildOuterBoundaryLink(startedge);
        }  
      }
    } // End of One Element Loop
  } // End of Main Loop
}

///////////////////////////////////////////////////////////////////////////////
// END                                                                       //
// Ptloc Preprocess Functions                                                //
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//  randomnation()   Generate a random number between 0 and `choices' - 1.   //
//                                                                           //
//  This is a simple linear congruential random number generator.  Hence, it //
//  is a bad random number generator, but good enough for most randomized    //
//  geometric algorithms.                                                    //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

unsigned long Randomnation(unsigned int choices)
{
  randomseed = (randomseed * 1366l + 150889l) % 714025l;
  return randomseed / (714025l / choices + 1);
}

///////////////////////////////////////////////////////////////////////////////
// Do Ptloc Functions                                                        //                        
// BEGIN                                                                     // 
/////////////////////////////////////////////////////////////////////////////// 

BOOL WalkThroughBoundary(CMeshNode& searchpoint, TriEdge* searchtri)
{
  TriEdge startedge;
  CMeshNode startp;
  CMeshNode *torg, *tdest;
  int orgorient, destorient;

  // Keep where we start search to startedge.
  startedge = *searchtri;
  // First we need form a ray that shot from searchtri to searchpoint.
  //   set startp as this ray's origin. Now, for simplicity, I just
  //   choose startedge's midpoint as startp. 
  torg = startedge.org();
  tdest = startedge.dest();
  startp.x = 0.5 * (torg->x + tdest->x);
  startp.y = 0.5 * (torg->y + tdest->y);
  
  // The WalkThroughBoundary Algorithm:
  //   Take startedge's next edge in boundary link as new searchtri. 
  //   Loop Until we encounter the startedge.
  //     Do current edge and ray intersection check.
  //     If they intersected then
  //       return TRUE. 
  //     else
  //       Take current edge's next edge in boundary link as new searchtri. 
  //       goto Loop.
  //   return FALSE.
  searchtri->nextboundaryself(); 
  while(*searchtri != startedge) {
    torg = searchtri->org();
    tdest = searchtri->dest();
    orgorient = Orient2D(startp, searchpoint, *torg);
    destorient = Orient2D(startp, searchpoint, *tdest);
    if(orgorient * destorient == 0) {
      if((orgorient == 0) && (destorient != 0)) {
        // startp, searchpoint, *torg collinear,
        if((FuzzyComp(torg->x, searchpoint.x) == 0) &&
           (FuzzyComp(torg->y, searchpoint.y) == 0)) {
          // This mean torg is just the searchpoint, we can return TRUE now.
          //   LocateResult should be eOnVertex.
          return TRUE; 	
        } else {   	
          // Check if `torg' is between `startp' and `searchpoint'.     		
          if (((startp.x < torg->x) == (torg->x < searchpoint.x)) && 
              ((startp.y < torg->y) == (torg->y < searchpoint.y))) {
            return TRUE;
          }
        }  
      } else if((orgorient != 0) && (destorient == 0)) {
        // startp, searchpoint, *tdest collinear,
        if((FuzzyComp(tdest->x, searchpoint.x) == 0) &&
           (FuzzyComp(tdest->y, searchpoint.y) == 0)) {
          // This mean tdest is just the searchpoint, we can return TRUE now.
          //   LocateResult should be eOnVertex.
          return TRUE; 	
        } else {   	
          //   Check if `tdest' is between `startp' and `searchpoint'.     		
          if (((startp.x < tdest->x) == (tdest->x < searchpoint.x)) && 
              ((startp.y < tdest->y) == (tdest->y < searchpoint.y))) {
            return TRUE;
          }
        }
      } else {
        // ray parallel with this edge, I don't care this edge.
      }	
    } else if(orgorient * destorient < 0) {
      orgorient = Orient2D(*torg, *tdest, startp);
      destorient = Orient2D(*torg, *tdest, searchpoint);
      if(orgorient * destorient == 0) {
        if(destorient == 0) {
          // This condition is searchpoint just locate on edge torg->tdest.
          //   LocateResult should be eOnEdge.
          return TRUE;
        } else if(orgorient == 0) {
        	// This condition should not happen. Because only the startedge
        	//   will hold this condition. Loop is already break before get 
        	//   here. 
          throw;
        }
      } else if(orgorient * destorient < 0) {
        return TRUE;
      }
    }	
    searchtri->nextboundaryself(); 
  }
  // We can't walk through this edge. return anyway.
  return FALSE;  
}

enum LocateResult PreciseLocate(CMeshNode& searchpoint, TriEdge* searchtri)
{
  TriEdge backtracetri;
  CMeshNode *forg, *fdest, *fapex;
  // CMeshNode *swappoint;
  int orgorient, destorient;
  int moveleft;
  
  // Where are we?
  forg = searchtri->org(); 
  fdest = searchtri->dest(); 
  fapex = searchtri->apex(); 
  while (1) {
    // Check whether the apex is the point we seek.
    if ((FuzzyComp(fapex->x, searchpoint.x) == 0) && 
        (FuzzyComp(fapex->y, searchpoint.y) == 0)) {
      searchtri->lprevself(); 
      recenttri = *searchtri;
      return eOnVertex;
    }
    // Does the point lie on the other side of the line defined by the
    //   triangle edge opposite the triangle's destination?
    destorient = Orient2D(*forg, *fapex, searchpoint);
    // Does the point lie on the other side of the line defined by the
    //   triangle edge opposite the triangle's origin?
    orgorient = Orient2D(*fapex, *fdest, searchpoint);
    if (destorient > 0) {
      if (orgorient > 0) {
        // Move left if the inner product of (fapex - searchpoint) and
        //   (fdest - forg) is positive.  This is equivalent to drawing
        //   a line perpendicular to the line (forg, fdest) passing
        //   through `fapex', and determining which side of this line
        //   `searchpoint' falls on.
        moveleft = (fapex->x - searchpoint.x) * (fdest->x - forg->x) +
                   (fapex->y - searchpoint.y) * (fdest->y - forg->y) > 0.0;
      } else {
        moveleft = 1;
      }
    } else {
      if (orgorient > 0) {
        moveleft = 0;
      } else {
        // The point we seek must be on the boundary of or inside this
        //   triangle.
        if (destorient == 0) {
          searchtri->lprevself(); 
          recenttri = *searchtri;
          return eOnEdge;
        }
        if (orgorient == 0) {
          searchtri->lnextself(); 
          recenttri = *searchtri;
          return eOnEdge; 
        }
        recenttri = *searchtri;
        return eInTriangle;
      }
    }

    // Move to another triangle.  Leave a trace `backtracktri' in case
    //   floating-point roundoff or some such bogey causes us to walk
    //   off a boundary of the triangulation.  We can just bounce off
    //   the boundary as if it were an elastic band.
    if (moveleft) {
      backtracetri = searchtri->lprev(); 
      fdest = fapex;
    } else {
      backtracetri = searchtri->lnext();
      forg = fapex;
    }
    *searchtri = backtracetri.sym(); 
    // Note: Here searchtri may be invalid. Let's check it.
    
    // Check for walking off the edge.
    if (!backtracetri.symexist()) {
      // Turn around.
      *searchtri = backtracetri; 
      // try to walk out the boundary to reach the other side boundary
      if(WalkThroughBoundary(searchpoint, searchtri)) {
        fapex = searchtri->apex(); 
      } else {
        // swappoint = forg;
        // forg = fdest;
        // fdest = swappoint;
        // fapex = searchtri->apex(); 
        // Check if the point really is beyond the triangulation boundary.
        // destorient = Orient2D(*forg, *fapex, searchpoint);
        // orgorient = Orient2D(*fapex, *fdest, searchpoint);
        // if ((orgorient < 0) && (destorient < 0)) {
        recenttri = *searchtri;	
        return eOutSide; 
        // }
      }
    } else {
      fapex = searchtri->apex(); 
    }
  }
}

// Used for the point location scheme of Mucke, Saias, and Zhu, to decide
//   how large a random sample of triangles to inspect.
#define SAMPLEFACTOR 11

enum LocateResult Locate(CMeshNode& searchpoint, TriEdge* searchtri)
{
  CMeshNode *torg, *tdest;
  double searchdist, dist;
  TriEdge sampletri;
  long samplenum;
  int ahead;
  int i;
  
  // Record the distance from the suggested starting triangle to the
  //   point we seek.
  if((searchtri->ele == INVALIDELEINDEX) || (searchtri->ele < 0) || 
     (searchtri->ele >= pmeshelem->GetSize()) ) {
    searchtri->ele = 0;
  } 
  torg = searchtri->org(); 
  searchdist = (searchpoint.x - torg->x) * (searchpoint.x - torg->x)
             + (searchpoint.y - torg->y) * (searchpoint.y - torg->y);
  // If a recently encountered triangle has been recorded and has not been
  //   deallocated, test it as a good starting point.
  if (recenttri.ele != INVALIDELEINDEX) {
    torg = recenttri.org(); 
    if ((FuzzyComp(torg->x, searchpoint.x) == 0) && 
        (FuzzyComp(torg->y, searchpoint.y) == 0)) {
      *searchtri = recenttri; 
      return eOnVertex;
    }
    dist = (searchpoint.x - torg->x) * (searchpoint.x - torg->x)
         + (searchpoint.y - torg->y) * (searchpoint.y - torg->y);
    if (dist < searchdist) {
      *searchtri = recenttri; 
      searchdist = dist;
    }
  } 
  
  // The number of random samples taken is proportional to the cube root of
  //   the number of triangles in the mesh.  The next bit of code assumes
  //   that the number of triangles increases monotonically.
  while (SAMPLEFACTOR * samples * samples * samples < pmeshelem->GetSize()) {
    samples++;
  }
  for(i = 0; i < samples; i ++) {
    samplenum = Randomnation(pmeshelem->GetSize());
    sampletri.ele = samplenum; 
    torg = sampletri.org(); 
    dist = (searchpoint.x - torg->x) * (searchpoint.x - torg->x)
         + (searchpoint.y - torg->y) * (searchpoint.y - torg->y);
    if (dist < searchdist) {
      *searchtri = sampletri; 
      searchdist = dist;
    }
  }
  
  // Here we need check if the searchtri is a boundary edge, If so, we need 
  //   get a non boundary edge as the searchtri, otherwise the sym() primitive 
  //   will cause exception. 
  if(!searchtri->symexist()) {
    TriEdge nextedge;
    nextedge = searchtri->lnext();
    while(!nextedge.symexist()) {
      nextedge.lnextself();
      if( nextedge == *searchtri) {
      	// It seems there only exist one elem.
      	return eOutSide;
      }	
    }
    *searchtri = nextedge;
  }
  
  // Where are we?
  torg = searchtri->org(); 
  tdest = searchtri->dest(); 
  // Check the starting triangle's vertices.
  if ((FuzzyComp(torg->x, searchpoint.x) == 0) && 
      (FuzzyComp(torg->y, searchpoint.y) == 0)) {
    recenttri = *searchtri;
    return eOnVertex;
  }
  if ((FuzzyComp(tdest->x, searchpoint.x) == 0) && 
      (FuzzyComp(tdest->y, searchpoint.y) == 0)) {
    searchtri->lnextself();
    recenttri = *searchtri;
    return eOnVertex;
  }
  // Orient `searchtri' to fit the preconditions of calling preciselocate().
  ahead = Orient2D(*torg, *tdest, searchpoint);
  if (ahead < 0) {
    // Turn around so that `searchpoint' is to the left of the
    //   edge specified by `searchtri'.
    searchtri->symself(); 
  } else if (ahead == 0) {
    // Check if `searchpoint' is between `torg' and `tdest'.
    if (((torg->x < searchpoint.x) == (searchpoint.x < tdest->x))
        && ((torg->y < searchpoint.y) == (searchpoint.y < tdest->y))) {
      recenttri = *searchtri;
      return eOnEdge;
    }
  }
  return PreciseLocate(searchpoint, searchtri);
}

/////////////////////////////////////////////////////////////////////////////// 
// END                                                                       //
// Do Ptloc Functions                                                        //                        
/////////////////////////////////////////////////////////////////////////////// 
